//**************************************************************************
//This file contains individual,population representation and genetic      *
//operators.(Programmed by NXH-version3.0-2002).                           *
//**************************************************************************
#ifndef __GENETIC__
#define __GENETIC__

#include<iostream.h>
#include<stdlib.h>
#include<string.h>
#include<assert.h>
#include<math.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
#include "dtree.h"
//Declaration for individual and population
struct _individual {
		      dnode *chrom;//chromosome of individual
		      int size;//size of individual
		      double branch;//branching of individual
		      double fitness;//fitness of individual
		   };
typedef struct _individual individual;
BOOL _cut=TRUE;
struct _pindividual{
		     individual iv;//individual
		     double pfitness;//fitness
		     double sfitness;//second objective value
		    // double knb;////kth-nearest neibour
		    };
typedef struct _pindividual pindividual;
pindividual *oldpop,*newpop;
//pindividual *oldarchive, *newarchive;
int poplen, alen; //population length
int generation,gen=0;
//some probabilities and statistics
double pcross,pinsert,pdelete,preplace,pmutate,prep,ptran;//probability for genetic operations
long ncross,ninsert,ndelete,nreplace,nmutate,nrep,ntran;//statistics
//individual bestsofar;
individual *bestcurrent;
double *averagefitness;
double *averagesfitness;
double *averagesize;
//*********************
//Copy an individual  *
//*********************
individual CopyIndividual(individual t,BOOL copychrom)
{
 individual v;
 if(copychrom)
  v.chrom=CopyDNode(t.chrom);
 else
  v.chrom=t.chrom;
  v.size=t.size;
  v.fitness=t.fitness;
  v.branch=t.branch;
 return v;
}
//*****************************************************
//Decode a individual to a derivation tree            *
//*****************************************************
stree Decode(individual v,BOOL stochastic)
{
   stree s;
   return DecodeDTree(v.chrom,stochastic);
}
//Display an individual
void DisplayIndividual(individual v)
{
 stree s;
 cout<<endl<<"Genotype Structure:";
 DisplayDTree(v.chrom);
 cout<<endl<<"Size:"<<v.size;
 cout<<endl<<"Branching:"<<v.branch;
 cout<<endl<<"Fitness:"<<v.fitness;
 s=Decode(v,FALSE);
 cout<<endl<<"Phenotype Structure:";
 Display(s);
 DeleteTree(s);
}
//*********************************************************
//Initialize a  shaped tree with a precified size(>=1)    *
//start tree=TRUE-start with alpha tree; FALSE- start with*
//beta tree; Supplied adjoining address (irrelevant when  *
//start with alpha tree                                   *
//*********************************************************
dnode *GenShapeTree(int &size,SIZE starttree,BYTE addno,BOOL shape)
{
 int j,asize;
 dnode *t;
  t=MakeDNode(starttree,addno);
  asize=1;
  if(shape) //Full tree
   {
     for(j=1;j<size;j++)
      if(AddFullBeta(t))
	 asize++;
    }
   else //random shape tree
    {
     for(j=1;j<size;j++)
       if(AddBeta(t))
	 asize++;
    }
   size=asize;
 return t;
}
//***************************
//Uniform Initialization    *
//***************************
void UniformInit(int minl,int maxl)
{
 int j,i;
 SIZE treeid;
 assert(minl>=MINSIZE);
 for(i=0;i<poplen;i++)
  {
   oldpop[i].iv.size=IRandom(minl,maxl);
   //choosing one alpha at random
   treeid=IRandom(1,numalpha);
    //find the alpha tree;
    j=0;
    while(treeid>0)
    {
     if(IsAlpha(etree[j]))
      treeid--;
     if(treeid>0)
      j++;
    }
   oldpop[i].iv.chrom=GenShapeTree(oldpop[i].iv.size,j,0,FALSE);
   oldpop[i].iv.branch=GetAVGDNode(oldpop[i].iv.chrom);
  }
}
//*****************************************************
//Ramp Initialization with a percentage of full tree  *
//*****************************************************
void RampInit(int minl, int maxl,double percentage)//(minl>=2)
{
 int i,j;
 SIZE treeid;
 assert(minl>=MINSIZE);
 for(i=0;i<poplen;i++)
  {
   oldpop[i].iv.size=IRandom(minl,maxl);
  //choosing one alpha at random
   treeid=IRandom(1,numalpha);
    //find the alpha tree;
    j=0;
    while(treeid>0)
    {
     if(IsAlpha(etree[j]))
      treeid--;
     if(treeid>0)
      j++;
    }
   if(Flip(percentage))
     oldpop[i].iv.chrom=GenShapeTree(oldpop[i].iv.size,j,0,TRUE);
   else
     oldpop[i].iv.chrom=GenShapeTree(oldpop[i].iv.size,j,0,FALSE);
   oldpop[i].iv.branch=GetAVGDNode(oldpop[i].iv.chrom);
  }
}
//Initialization Procedure
void Initialization(int minl,int maxl,int type,double percentage)
{
 if (type==0)
  UniformInit(minl,maxl);
 else
  RampInit(minl,maxl,percentage);
}
//*************************************************
//Selections works on Normalized Fitness          *
//*************************************************
//Propotionate Selection
int ProSelect()
{
  double k=Next_Double();
  int i=0;
  while((i<poplen-1)&&(k>0))
   {
     k-=oldpop[i].iv.fitness;
     if(k>0)
      i++;
   }
   return i;
}
//Tournement Selection
int TourSelect(int size)
{
   int i,j,pos=0;
   double max=-1000;
   assert(size<=MAXTOUR);
   for(i=0;i<size;i++)
    {
      j=IRandom(0,poplen-1);
      if (oldpop[j].iv.fitness>max)
       {
	 max=oldpop[j].iv.fitness;
	 pos=j;
       }
    }
   return pos;
}
//Linear Ranking Selection
//**************************************************
//Crossover Section                                *
//**************************************************
BOOL RootAdjoinability (SIZE treeid)
{
 address *p;
 p=adtable[treeid].head;
 if(p)
 {
  while(p->next) p=p->next;
  if(p->addr==etree[treeid].root)
     return TRUE;
 }
return FALSE;
}

//***********************************
//Check if a node is root adjoined  *
//***********************************
BOOL RootAdjoined(dnode *p)
{
 dnode *q;
 address *ad;
 BYTE i;
 q=p->children;
 if(q)
 {
  while (q->sibling)
   q=q->sibling;
  ad=adtable[p->id].head;
  i=0;
  while(i<q->addno)
   {
    ad=ad->next;
    i++;
   }
  if(ad->addr==etree[p->id].root)
      return TRUE;
 }

 return FALSE;
}
//*******************************
//Crossover point distribution  *
//******************************
double *_pointdis;
int  _dnodeindex1;
void DNodeProb(dnode *t)
{
 dnode *q;
//cout<<"Hi, there from DNodeProb"<<endl;
 _pointdis[_dnodeindex1]=1.0;
 //check if the node has an root adjunction
 if(RootAdjoined(t)) //the root address
     _pointdis[_dnodeindex1]++;
//cout<<"Hi, There from DNodeProb after calling RootAdjoined Node number:"<<_dnodeindex1<<endl;
 _dnodeindex1++;
}
int GetCrossPoint(individual p)
{
 int i,pos;
 double s;
//cout<<"size of chromosome:"<<p.size<<endl;
_pointdis=new double[p.size]; assert(_pointdis!=NULL);
 _dnodeindex1=0;
 _doitdnode=DNodeProb;
 PreorderDNode(p.chrom);
//normalise the distribution exclude the root
 s=0;
 for(i=1;i<p.size;i++)
  s+=_pointdis[i];
 for(i=1;i<p.size;i++)
  _pointdis[i]=_pointdis[i]/s;
//choose the crossover point
 s=Next_Double();
 pos=1;
 while ((pos<p.size-1)&&(s>0))
 {
   s-=_pointdis[pos];
   if(s>0)
    pos++;
 }
 delete [] _pointdis;
 return pos+1;
}

//****************************************************************
//This function implement subtree crossover by swaping two       *
//subtrees. The adjoing addresses (contexts) are maintained      *
//****************************************************************
BOOL SubTreeSwap(individual parent1,individual parent2,individual &child1,individual &child2,BOOL fairsize,int diff)
{
  int crosspoint1,crosspoint2,j,size1,size2,temp1;
  dnode *dnodeparent1,*dnodeparent2,*dnode1,*dnode2,*temp2,*ch,*ch2;
  BOOL found1,found2,act,exroot1,exroot2,root1,root2;
  address *p;
  snode   *t;
  BYTE i;
  llist l;
  j=0;
  while(j<MAXATEMPT)
  {

// cout<< "hi, before getting crosspoints"<<endl;
   crosspoint1=GetCrossPoint(parent1);
   crosspoint2=GetCrossPoint(parent2);

//cout<<"Chrom size:"<<parent1.size<<" Crosspoint:"<<crosspoint1<<endl;
//cout<<"Chrom size:"<<parent2.size<<" Crosspoint:"<<crosspoint2<<endl;


   LocateDNode(crosspoint1,parent1.chrom,NULL);
   dnode1=_idnode;
   dnodeparent1=_idnodeparent;
   LocateDNode(crosspoint2,parent2.chrom,NULL);
   dnode2=_idnode;
   dnodeparent2=_idnodeparent;
//check if they have the same tree X_Beta root
//by checking the adtable (because there might be constraints in future)
   p=adtable[dnodeparent1->id].head;
   i=0;
   while(i<dnode1->addno)
    {
     p=p->next;
     i++;
    }
    t=(p->source).head;
    found1=FALSE;
    while(t && !found1)
      if(t->id==dnode2->id)
       found1=TRUE;
      else
       t=t->next;
    if(found1)
    {
      p=adtable[dnodeparent2->id].head;
      i=0;
     while(i<dnode2->addno)
       {
	p=p->next;
	i++;
       }
    t=(p->source).head;
    found2=FALSE;
    while(t && !found2)
      if(t->id==dnode1->id)
       found2=TRUE;
      else
       t=t->next;
      if(found2)//they are compatible
       {
	act=TRUE;

	exroot1=FALSE;
	exroot2=FALSE;
	root1=RootAdjoined(dnode1);
	root2=RootAdjoined(dnode2);
	if(root1)
	    if(Flip(0.5))
	       exroot1=TRUE;
	if(root2)
	    if(Flip(0.5))
	       exroot2=TRUE;
       //can't swap if both have root and one don't swap root, one swap root
	if(exroot1 && root2 && !exroot2)
	    act=FALSE;
	if(exroot2 && root1 && !exroot1)
	    act=FALSE;
       //canculate the size of the two sub-trees
	  size1=GetSizeDNode(dnode1,TRUE);
	  size2=GetSizeDNode(dnode2,TRUE);

	if(exroot1)
	  {
	    ch=dnode1->children;
	    while(ch->sibling) ch=ch->sibling;
	    size1=size1-GetSizeDNode(ch,TRUE);
	  }
	if(exroot2)
	 {
	  ch=dnode2->children;
	  while(ch->sibling) ch=ch->sibling;
	  size2=size2-GetSizeDNode(ch,TRUE);
	 }

      if(fairsize)//fairsize crossover
	  //check if the difference is small enough
	  if(abs(size1-size2)>diff)
	   act=FALSE;
       //check the size limit
	if((parent1.size-size1+size2<MINSIZE)||(parent2.size-size2+size1<MINSIZE))
	 act=FALSE;
	if(((parent1.size-size1+size2>MAXSIZE)||(parent2.size-size2+size1>MAXSIZE))&& !_cut)
	   act=FALSE;
	if(act) //swaping
	 {
	 //swaping tree id first
	 temp1=dnode1->id;
	 dnode1->id=dnode2->id;
	 dnode2->id=temp1;
	 //then swaping lexlist
	 l=dnode1->lexlist;
	 dnode1->lexlist=dnode2->lexlist;
	 dnode2->lexlist=l;
	 //then swaping the children
	 temp2=dnode1->children;
	 dnode1->children=dnode2->children;
	 dnode2->children=temp2;

	 if(exroot1&&exroot2)
	  {

	   //returning swap of  the root adjunctions
	    ch=dnode1->children;
	    while(ch->sibling) ch=ch->sibling;
	    ch2=dnode2->children;
	    while (ch2->sibling) ch2=ch2->sibling;
	  //swaping back the root adjoining node
	    temp1=ch->id;
	    ch->id=ch2->id;
	    ch2->id=temp1;
	    l=ch->lexlist;
	    ch->lexlist=ch2->lexlist;
	    ch2->lexlist=l;
	    temp2=ch->children;
	    ch->children=ch2->children;
	    ch2->children=temp2;
	  }
	 else
	  if(exroot1)
	   {
	     //root adjunction at dnode1 and no root adjunction in dnode2 before swaping
	    if(RootAdjoinability(dnode1->id))
	    {
	     ch2=dnode2->children;
	     while(ch2->sibling) ch2=ch2->sibling;
	     ch=dnode1->children;
	     if(ch)
	      {

		   while(ch->sibling) ch=ch->sibling;
		   ch->sibling=ch2;
		   ch2->addno=adtable[dnode1->id].card-1;
	       }
	     else
	       {
		    dnode1->children=ch2;
		    ch2->addno=adtable[dnode1->id].card-1;
	       }
	   //cutting the old root
	    ch=dnode2->children;
	    if(ch==ch2)
	     dnode2->children=NULL;
	    else
	     {
	      while (ch->sibling!=ch2) ch=ch->sibling;
	      ch->sibling=NULL;
	     }
	    }

	   }
	  else
	   if(exroot2)
	    {
	    //root adjunction at dnode2 and no root adjunction at dnode1 before swaping
	     if(RootAdjoinability(dnode2->id))
	      {
	       ch=dnode1->children;
	       while(ch->sibling) ch=ch->sibling;
	       ch2=dnode2->children;
	       if(ch2)
		{
		 while(ch2->sibling) ch2=ch2->sibling;
		 ch2->sibling=ch;
		 ch->addno=adtable[dnode2->id].card-1;
		}
	       else
		{
		 dnode2->children=ch;
		 ch->addno=adtable[dnode2->id].card-1;
		}
	     //cutting the root
	       ch2=dnode1->children;
	       if(ch2==ch)
		dnode1->children=NULL;
	       else
	       {
		while(ch2->sibling!=ch) ch2=ch2->sibling;
		ch2->sibling=NULL;
	       }
	      }

	    }

	 //recompute size and branch
	  child1.chrom=parent1.chrom;
	  child2.chrom=parent2.chrom;
	  child1.size=parent1.size-size1+size2;
	  if(child1.size>MAXSIZE) //randomly cut
	   {
	    SIZE k;
	    for(k=MAXSIZE;k<child1.size;k++)
	      DeleteBeta(dnode1);
	     child1.size=MAXSIZE;
	    }
	  child1.branch=GetAVGDNode(child1.chrom);
	  child2.size=parent2.size-size2+size1;
	  if(child2.size>MAXSIZE) //randomly cut
	   {
	    SIZE k;
	    for(k=MAXSIZE;k<child2.size;k++)
	      DeleteBeta(dnode2);
	    child2.size=MAXSIZE;
	    }
	    child2.branch=GetAVGDNode(child2.chrom);
	    return TRUE;
	  }
	}
      }
   j++; //Next attempt
  }
return FALSE;
}
//****************************************************************
//This function implement subtree crossover by swaping two       *
//subtrees. The adjoing addresses (contexts) are maintained      *
//****************************************************************
BOOL OldSubTreeSwap(individual parent1,individual parent2,individual &child1,individual &child2,BOOL fairsize,int diff)
{
  int crosspoint1,crosspoint2,j,size1,size2,temp1;
  dnode *dnodeparent1,*dnodeparent2,*dnode1,*dnode2,*temp2;
  BOOL found1,found2,act;
  address *p;
  snode   *t;
  BYTE i;
  llist l;
  j=0;
  while(j<MAXATEMPT)
  {
   crosspoint1=IRandom(2,parent1.size);//exclude the root
   crosspoint2=IRandom(2,parent2.size);//exclude the root
   LocateDNode(crosspoint1,parent1.chrom,NULL);
   dnode1=_idnode;
   dnodeparent1=_idnodeparent;
   LocateDNode(crosspoint2,parent2.chrom,NULL);
   dnode2=_idnode;
   dnodeparent2=_idnodeparent;
//check if they have the same tree X_Beta root
//by checking the adtable (because there might be constraints in future)
   p=adtable[dnodeparent1->id].head;
   i=0;
   while(i<dnode1->addno)
    {
     p=p->next;
     i++;
    }
    t=(p->source).head;
    found1=FALSE;
    while(t && !found1)
      if(t->id==dnode2->id)
       found1=TRUE;
      else
       t=t->next;
    if(found1)
    {
      p=adtable[dnodeparent2->id].head;
      i=0;
     while(i<dnode2->addno)
       {
	p=p->next;
	i++;
       }
    t=(p->source).head;
    found2=FALSE;
    while(t && !found2)
      if(t->id==dnode1->id)
       found2=TRUE;
      else
       t=t->next;
      if(found2)//they are compatible
       {
	act=TRUE;
       //canculate the size of the two sub-trees
	  size1=GetSizeDNode(dnode1,TRUE);
	  size2=GetSizeDNode(dnode2,TRUE);
	if(fairsize)//fairsize crossover
	  //check if the difference is small enough
	  if(abs(size1-size2)>diff)
	   act=FALSE;
       //check the size limit
	if((parent1.size-size1+size2<MINSIZE)||(parent2.size-size2+size1<MINSIZE))
	 act=FALSE;
	if(((parent1.size-size1+size2>MAXSIZE)||(parent2.size-size2+size1>MAXSIZE))&& !_cut)
	   act=FALSE;
	if(act) //swaping
	 {
	 //swaping tree id first
	 temp1=dnode1->id;
	 dnode1->id=dnode2->id;
	 dnode2->id=temp1;
	 //then swaping lexlist
	 l=dnode1->lexlist;
	 dnode1->lexlist=dnode2->lexlist;
	 dnode2->lexlist=l;
	 //then swaping the children
	 temp2=dnode1->children;
	 dnode1->children=dnode2->children;
	 dnode2->children=temp2;
	 //recompute size and branch
	  child1.chrom=parent1.chrom;
	  child2.chrom=parent2.chrom;
	  child1.size=parent1.size-size1+size2;
	  if(child1.size>MAXSIZE) //randomly cut
	   {
	    SIZE k;
	    for(k=MAXSIZE;k<child1.size;k++)
	      DeleteBeta(dnode1);
	     child1.size=MAXSIZE;
	    }
	  child1.branch=GetAVGDNode(child1.chrom);
	  child2.size=parent2.size-size2+size1;
	  if(child2.size>MAXSIZE) //randomly cut
	   {
	    SIZE k;
	    for(k=MAXSIZE;k<child2.size;k++)
	      DeleteBeta(dnode2);
	    child2.size=MAXSIZE;
	    }
	    child2.branch=GetAVGDNode(child2.chrom);
	    return TRUE;
	  }
	}
      }
   j++; //Next attempt
  }
return FALSE;
}
//****************************************************************
//This function implement subtree crossover by swaping two       *
//subtrees. The adjoing addresses (contexts) are not maintained  *
//****************************************************************
//BOOL SubTreeSwap2(int parent1,int parent2,individual *child1,individual *child2,BOOL fairsize,int diff)
//{
//}
//****************************************************************
//This function implement subtree crossover by swaping children  *
// of two nodes. The adjoing addresses (contexts) are maintained *
//****************************************************************
BOOL SubSubTreeSwap(individual parent1,individual parent2,individual &child1,individual &child2,BOOL fairsize,int diff)
{
  int crosspoint1,crosspoint2,j,size1,size2;
  dnode *dnode1,*dnode2,*q,*temp;
  BOOL found1,found2;
  address *p;
  snode   *t;
  BYTE i;
  j=0;
  while (j<=MAXATEMPT)
  {
   crosspoint1=IRandom(1,parent1.size);//include the root
   crosspoint2=IRandom(1,parent2.size);//include the root
   LocateDNode(crosspoint1,parent1.chrom,NULL);
   dnode1=_idnode;
   LocateDNode(crosspoint2,parent2.chrom,NULL);
   dnode2=_idnode;
   found1=TRUE;
   q=dnode2->children;
   if(q) //check whether children nodes of dnode2 can be adjoined into dnode1
    {
     while (q&&found1)
      {
       if(q->addno>=adtable[dnode1->id].card)
	found1=FALSE;
       else
	{
	 p=adtable[dnode1->id].head;
	 i=0;
	 //find the address
	 while(i<q->addno)
	 {
	  p=p->next;
	  i++;
	 }
	 //find the tree
	  t=(p->source).head;
	  found2=FALSE;
	 while (t && !found2)
	   if(t->id==q->id)
	    found2=TRUE;
	   else
	  t=t->next;
	  if (!found2)
	    found1=FALSE;
	}
       q=q->sibling;
      }
    }
    if(found1)
     {
      //check if children nodes of dnode1 can be adjoined into dnode2
      q=dnode1->children;
     if(q)
      {
	while (q&&found1)
	 {
	  if(q->addno>=adtable[dnode2->id].card)
	   found1=FALSE;
	  else
	   {
	    p=adtable[dnode2->id].head;
	    i=0;
	  //find the address
	   while(i<q->addno)
	    {
	      p=p->next;
	      i++;
	    }
	  //find the tree
	   t=(p->source).head;
	   found2=FALSE;
	  while (t && !found2)
	   if(t->id==q->id)
	    found2=TRUE;
	   else
	    t=t->next;
	   if (!found2)
	    found1=FALSE;
	  }
	  q=q->sibling;
	 }
	}
     }
   if (found1)
   {
     //check size
     size1=GetSizeDNode(dnode1,TRUE);
     size2=GetSizeDNode(dnode2,TRUE);
     if(fairsize)//if fair size crossover
      if(abs(size1-size2)>diff)
      found1=FALSE;
  if((parent1.size-size1+size2<MINSIZE)||(parent2.size-size2+size1<MINSIZE))
      found1=FALSE;
  if(((parent1.size-size1+size2>MAXSIZE)||(parent2.size-size2+size1))&&!_cut)
       found1=FALSE;
     if(found1)
      {
       temp=dnode1->children;
       dnode1->children=dnode2->children;
       dnode2->children=temp;
       child1.chrom=parent1.chrom;
       child2.chrom=parent2.chrom;
       child1.size=parent1.size-size1+size2;
       if(child1.size>MAXSIZE)//randomly cut
	{
	 SIZE k;
	 for(k=MAXSIZE;k<child1.size;k++)
	  DeleteBeta(dnode1);
	 child1.size=MAXSIZE;
	}
       child1.branch=GetAVGDNode(child1.chrom);
       child2.size=parent2.size-size2+size1;
      if(child2.size>MAXSIZE)//randomly cut
	{
	 SIZE k;
	 for(k=MAXSIZE;k<child2.size;k++)
	  DeleteBeta(dnode2);
	 child2.size=MAXSIZE;
	}
       child2.branch=GetAVGDNode(child2.chrom);
	 return TRUE;
       }
    }
   j++;
  }
return FALSE;
}
//******************************************************************
//This function implement subtree crossover by swaping children    *
//of two nodes. The adjoing addresses (contexts) are notmaintained *
//******************************************************************
//BOOL SubSubTreeSwap2(int parent1,int parent2,individual *child1,individual *child2,BOOL fairsize,int diff)
//{
//}
//**************************************************
//Mutation Section                                 *
//**************************************************************
//This function will replace subtree by randomly               *
//generated one; fairsize=TRUE the difference in size must be  *
//between 0 and diff; shape=TRUE means FULL otherwise RANDOM   *
//use size to specidied the size you want                      *
//The address (context) is main tained                         *
//**************************************************************
BOOL ReplaceSubTree(individual &child,int size, BOOL shape,BOOL fairsize,int diff)
{
  int replacepoint,j,size1,size2,k,h,l,treeid;
  dnode *parentdnode1,*dnode1,*dnode2,*temp,*q;
  address *p;
  BYTE i;
  j=0;
  while(j<MAXATEMPT)
  {
   replacepoint=IRandom(1,child.size);//include the root
   LocateDNode(replacepoint,child.chrom,NULL);
   dnode1=_idnode;
   parentdnode1=_idnodeparent;
   size2=GetSizeDNode(dnode1,TRUE);
   if(!fairsize)
    size1=size;
   else
   if(Flip(0.5))
     size1=size2+IRandom(0,diff);//increase size or decrease size
   else
    {
       k=IRandom(0,diff);
     if(size2>k)
       size1=size2-k;
     else
       size1=size2-IRandom(0,size2);
    }
   if(replacepoint==1)//replace at root the generate totally new tree
     {
	if(((child.size-size2+size1<=MAXSIZE)||_cut)&&(child.size-size2+size1>=MINSIZE))
	  {
	   temp=child.chrom;
	   //choosing one alpha at random
	   treeid=IRandom(1,numalpha);
	   //find the alpha tree;
	   l=0;
	   while(treeid>0)
	    {
	      if(IsAlpha(etree[l]))
	       treeid--;
	      if(treeid>0)
		 l++;
	     }
	  child.chrom=GenShapeTree(size1,l,0,shape);
	  DeleteDNode(temp);
	  child.size=size1;
	  if(child.size>MAXSIZE)//randomly cut
	   {
	    for(l=MAXSIZE;l<child.size;l++)
	     DeleteBeta(child.chrom);
	    child.size=MAXSIZE;
	   }
	  child.branch=GetAVGDNode(child.chrom);
	  return TRUE;
       }
     }
    else//not the root
     if(((child.size-size2+size1<=MAXSIZE)||_cut)&&(child.size-size2+size1>=MINSIZE))
      {
      //find the replace tree for dnode
      //find the address
      p=adtable[parentdnode1->id].head;
      i=0;
      while(i<dnode1->addno)
       {
	p=p->next;
	i++;
       }
      //randomly choose a tree
      k=GetTreeID(p->source);

      q=dnode1->children;
      //adding beta trees
      dnode2=GenShapeTree(size1,k,dnode1->addno,shape);
      dnode1->id=dnode2->id;
      dnode1->children=dnode2->children;
      dnode1->lexlist=dnode2->lexlist;
      delete dnode2;
      child.size=child.size-size2+size1;
      if(child.size>MAXSIZE)//randomly cut
	  {
	   for(l=MAXSIZE;l<child.size;l++)
	    DeleteBeta(dnode1);
	   child.size=MAXSIZE;
	  }
      child.branch=GetAVGDNode(child.chrom);
	//Delete old chidren of dnode1;
	while(q)
	 {
	   temp=q;
	   q=q->sibling;
	   DeleteDNode(temp);
	  }
	 return TRUE;
	}
   j++;
  }
return FALSE;
}
//**************************************************************
//This function will replace subtree by randomly               *
//generated one; fairsize=TRUE the difference in size must be  *
//between 0 and diff; shape=TRUE means FULL otherwise RANDOM   *
//use size to specidied the size you want                      *
//The address (context) is not maintained                      *
//**************************************************************
//BOOL ReplaceSubTree(individual &child,int size, BOOL shape,BOOL fairsize,int diff)
//{
//}
//*********************************************************
//this function replace a node by another compatible node *
//*********************************************************
BOOL PointReplace(individual &child)
{
 int j,k,h;
 SIZE i;
 dnode *q,*dnode1,*parentdnode1;
 address *p;
 snode *t;
 int replacepoint;
 BOOL found,*check,found1;
 check=new BOOL[numetree];
 assert(check!=NULL);
 j=0;
 while(j<MAXATEMPT)
  {
   replacepoint=IRandom(1,child.size);
   LocateDNode(replacepoint,child.chrom,NULL);
   dnode1=_idnode;
   parentdnode1=_idnodeparent;
   for(i=0;i<numetree;i++)
    check[i]=FALSE;
    k=0;
   for(i=0;(i<numetree)&&(i!=dnode1->id);i++)
    {
     found=FALSE;
    //the check compatibility with parent address
     if(parentdnode1)
      {
       p=adtable[parentdnode1->id].head;
       h=0;
       while (h<dnode1->addno)
	{
	 p=p->next;
	 h++;
	}
       t=(p->source).head;
       found1=FALSE;
       while (t && !found1)
	if (t->id==i)
	 found1=TRUE;
	else
	 t=t->next;
       if(!found1)
	found=TRUE;
      }
     q=dnode1->children;
     while(q && !found)
      {
       //check if tree i has address and tree q->id
      if(q->addno>=adtable[i].card)//no such address
	 found=TRUE;
       else//check if q->id is in the source list
	{
      //find the address
	  p=adtable[i].head;
	  h=0;
	  while(p&&(h<q->addno))
	   {
	     p=p->next;
	     h++;
	   }
	  t=(p->source).head;
	  found1=FALSE;
	  while(t&&!found1)
	   if(t->id==q->id)
	    found1=TRUE;
	   else
	    t=t->next;
	  if(!found1)
	   found=TRUE;
	}
       q=q->sibling;
      }
     if(!found)//this tree is a candidate
     {
      check[i]=TRUE;
      k++;
     }
    }
   if(k>0)//there are some trees to replace
    {
      k=IRandom(1,k); //choose at random
      h=0;
      while(k>0)
      {
       if(check[h])
	k--;
       if(k>0)
	h++;
      }
      dnode1->id=h;
      DeleteLList(dnode1->lexlist);
      dnode1->lexlist=BuildLList(etree[h].root);
     //deletele list
     delete [] check;
     return TRUE;
    }
   j++;
  }
return FALSE;
}
//***************************
//Directed Point Mutation   *
//***************************
SIZE *_replacelist;
BOOL DirectedPointReplace(individual &child)
{
 int j;
 dnode *dnode1;
 int replacepoint;
 SIZE temp;
 j=0;
 while(j<MAXATEMPT)
  {
   replacepoint=IRandom(1,child.size);
   LocateDNode(replacepoint,child.chrom,NULL);
   dnode1=_idnode;
   if(_replacelist[dnode1->id]>=0)
     {
      temp=_replacelist[dnode1->id];
      dnode1->id=temp;
      DeleteLList(dnode1->lexlist);
      dnode1->lexlist=BuildLList(etree[temp].root);
      return TRUE;
     }
   j++;
  }
return FALSE;
}
//**************************
//      Lexicon Mutation   *
//**************************
void DNodeLexMutation(dnode *t)
{
 if(t)
 {
  if(!EmptyLList(t->lexlist))
  {
   DeleteLList(t->lexlist);
   t->lexlist=BuildLList(etree[t->id].root);
  }
 }
}
//**************************
//Point Lexicon Mutation   *
//**************************
BOOL PointLexMutation(individual &child)
{
 dnode *dnode1;
 int replacepoint;
 SIZE temp;
  if(child.chrom)
  {
    replacepoint=IRandom(1,child.size);
    LocateDNode(replacepoint,child.chrom,NULL);
    dnode1=_idnode;
    DNodeLexMutation(dnode1);
    return TRUE;
   }
  return FALSE;
}
//******************
//Lexicon Mutation *
//******************
BOOL LexMutation(individual &child)
{
 _doitdnode=DNodeLexMutation;
 PreorderDNode(child.chrom);
 return TRUE;
}

//*****************
//SIZE MUTATIONS  *
//*****************
BOOL Delete(individual &child,SIZE size)
{
 SIZE i;
 if(child.size-size>=MINSIZE)
  {
   for(i=1;i<=size;i++)
    if( DeleteBeta(child.chrom))
      child.size--;
    child.branch=GetAVGDNode(child.chrom);
   return TRUE;
  }
 return FALSE;
}
BOOL Insert(individual &child,SIZE size,BOOL shape)
{
 SIZE i;
 BOOL b;
 if(child.size+size<=MAXSIZE)
  {
   for(i=1;i<=size;i++)
    {
     if(shape)
      b=AddFullBeta(child.chrom);
     else
      b=AddBeta(child.chrom);
   if(b)
    child.size++;
  }
    child.branch=GetAVGDNode(child.chrom);
    return TRUE;
  }
 return FALSE;
}
//******************************************************************
//One Step Mutation: either change value or add or delete one node *
//******************************************************************
BOOL OneStepMutation(individual &child,BOOL shape)
{
 int action;
 action=IRandom(1,4);
 switch(action)
  {
   case 1:
      if(PointReplace(child))
	return TRUE;
       else
	return FALSE;
   case 2:
      if (Delete(child,1))
	  return TRUE;
      else
	return FALSE;
   case 3:
      if (Insert(child,1,shape))
	 return TRUE;
      else
       return FALSE;
   case 4:
    if(PointLexMutation(child))
     return TRUE;
    else
     return FALSE;
  }
 return FALSE;
}
//*****************************************************
//Relocation, Transposition and Replication Operators *
//*****************************************************
//****************************
//relocation of a subtree    *
//****************************
BOOL Relocate(individual &child)
{
 dnode *dnode1,*parentdnode1,*dnode2,*p,*q;
 int j,relocatepoint,opendnodeno;
 BYTE i,k;
 BOOL *check,found,act;
 address *a;
 snode   *t;
  _opendnodeno=0;
  _countorfind=TRUE;//counting
  _doitdnode=CountOpenDNode;
   PreorderDNode(child.chrom);//result is in _opendnodeno
  if(_opendnodeno<=0)
   return FALSE;
  else
   opendnodeno=_opendnodeno;
 j=0;
 while (j<MAXATEMPT)
  {
   relocatepoint=IRandom(2,child.size);//exclude the root node
   LocateDNode(relocatepoint,child.chrom,NULL);
   dnode1=_idnode;
   parentdnode1=_idnodeparent;
   //find open node
     _opendnodeno=IRandom(1,opendnodeno);
     _countorfind=FALSE;//finding
     _opendnode=NULL;
     _doitdnode=CountOpenDNode;
     PreorderDNode(child.chrom);//results is in _opendnode
     dnode2=_opendnode;
     if(!InDNode(dnode2,dnode1))
      {
       //check if dnode1 can be adjoined into dnode2
       //find open addresses

       check=new BOOL[adtable[dnode2->id].card];
       assert(check!=NULL);
       for(i=0;i<adtable[dnode2->id].card;i++)
	check[i]=FALSE;
       act=FALSE;
       p=dnode2->children;
       while(p)
	{
	 check[p->addno]=TRUE;
	 p=p->sibling;
	}
       k=0;
       a=adtable[dnode2->id].head;
       for(i=0;i<adtable[dnode2->id].card;i++)
       {
	if(!check[i])
	 {
	  //check if dnode1 can adjoint into this address
	  found=FALSE;
	  t=(a->source).head;
	  while(t&&!found)
	    if(t->id==dnode1->id)
	     found=TRUE;
	    else
	     t=t->next;
	  if(found)
	   k++;
	  else
	   check[i]=TRUE;
	 }
	 a=a->next;
       }
     //choose the address to attach
       if(k>0)
       {
	 k=IRandom(1,k);
	 i=0;
	 while(k>0)
	  {
	   if(!check[i])
	    k--;
	   if(k>0)
	    i++;
	  }
	dnode1->addno=i;
	act=TRUE;//action
       //Deattach dnode1
	 p=parentdnode1->children;
	 if(p==dnode1)
	  parentdnode1->children=p->sibling;
	 else
	  {
	   while(p->sibling!=dnode1)
	    p=p->sibling;
	   p->sibling=dnode1->sibling;
	  }
       //Relocate dnode1
	  p=dnode2->children;
	  if(p==NULL)
	    {
	      dnode2->children=dnode1;
	      dnode1->sibling=NULL;
	    }
	  else
	   {
	    found=FALSE;
	    while(p&&!found)
	      if(p->addno>dnode1->addno)
		 found=TRUE;
	      else
	       p=p->sibling;
	    if(found)
	      {
		q=dnode2->children;
		if(q==p)//find in first place
		 {
		   dnode1->sibling=q;
		   dnode2->children=dnode1;
		 }
		else
		 {
		  while(q->sibling!=p)
		     q=q->sibling;
		  q->sibling=dnode1;
		  dnode1->sibling=p;
		 }
	      }
	     else//insert at the end
	       {
		 q=dnode2->children;
		 while(q->sibling)
		  q=q->sibling;
		 q->sibling=dnode1;
		 dnode1->sibling=NULL;
	       }
	    }
	  }
	 delete [] check;
	 if(act)
	 {
	  //recompute the branch factor
	  child.branch=GetAVGDNode(child.chrom);
	  return TRUE;
	 }
       }
   j++;
  }
 return FALSE;
}
//Replication
BOOL Replicate(individual &child,BOOL recursive)
{
 dnode *dnode1,*dnode2,*p,*q,*temp;
 int j,relocatepoint,size1,h,opendnodeno;
 BYTE i,k;
 BOOL *check,found,act;
 address *a;
 snode   *t;
 //find open node
  _opendnodeno=0;
  _countorfind=TRUE;//counting
  _doitdnode=CountOpenDNode;
   PreorderDNode(child.chrom);//result is in _opendnodeno
   if(_opendnodeno<=0)//if there is no open nodes
       return FALSE;
   else
    opendnodeno=_opendnodeno;
 j=0;
 while (j<MAXATEMPT)
  {
   relocatepoint=IRandom(2,child.size);//exclude the root node
   LocateDNode(relocatepoint,child.chrom,NULL);
   dnode1=_idnode;
   size1=GetSizeDNode(dnode1,TRUE);
  if((child.size+size1<=MAXSIZE )||_cut)
   {
     //find an open node
     _opendnodeno=IRandom(1,opendnodeno);
     _countorfind=FALSE;//finding
     _opendnode=NULL;
     _doitdnode=CountOpenDNode;
     PreorderDNode(child.chrom);//results is in _opendnode
     dnode2=_opendnode;
    if(recursive || !InDNode(dnode2,dnode1))
      {
       //check if dnode1 can be adjoined into dnode2
       //find open addresses
       act=FALSE;
       check=new BOOL[adtable[dnode2->id].card];
       assert(check!=NULL);
       for(i=0;i<adtable[dnode2->id].card;i++)
	check[i]=FALSE;
       p=dnode2->children;
       while(p)
	{
	 check[p->addno]=TRUE;
	 p=p->sibling;
	}
       k=0;
       a=adtable[dnode2->id].head;
       for(i=0;i<adtable[dnode2->id].card;i++)
       {
	if(!check[i])
	 {
	  //check if dnode1 can adjoint into this address
	  found=FALSE;
	  t=(a->source).head;
	  while(t&&!found)
	    if(t->id==dnode1->id)
	     found=TRUE;
	    else
	     t=t->next;
	  if(found)
	   k++;
	  else
	   check[i]=TRUE;
	 }
	a=a->next;
       }
     //choose the address to attach
       if(k>0)
       {
	 k=IRandom(1,k);
	 i=0;
	 while(k>0)
	  {
	   if(!check[i])
	    k--;
	   if(k>0)
	    i++;
	  }
	act=TRUE;//action
	//copy dnode1
	temp=dnode1;
	dnode1=CopyDNode(temp);
	dnode1->addno=i;
	dnode1->sibling=NULL;
       //Replicate dnode1
	  p=dnode2->children;
	  if(p==NULL)
	    dnode2->children=dnode1;
	  else
	   {
	    found=FALSE;
	    while(p&&!found)
	      if(p->addno>dnode1->addno)
		 found=TRUE;
	      else
	       p=p->sibling;
	    if(found)
	      {
		q=dnode2->children;
		if(q==p)//find in first place
		 {
		   dnode1->sibling=q;
		   dnode2->children=dnode1;
		 }
		else
		 {
		  while(q->sibling!=p)
		     q=q->sibling;
		  q->sibling=dnode1;
		  dnode1->sibling=p;
		 }
	      }
	     else//insert at the end
	       {
		 q=dnode2->children;
		 while(q->sibling)
		  q=q->sibling;
		 q->sibling=dnode1;
	       }
	    }
	  }
	 delete [] check;
	 if(act)
	 {
	  //check size and cut
	  child.size+=size1;
	  if(child.size>MAXSIZE)
	   {
	    for(h=MAXSIZE;h<child.size;h++)
	     DeleteBeta(child.chrom);
	    child.size=MAXSIZE;
	   }
	  //recompute the branch factor
	  child.branch=GetAVGDNode(child.chrom);
	  return TRUE;
	 }
       }
    }
   j++;
  }
 return FALSE;
}
#endif
