//************************************************************************
//This module contains fitness case and functions to compute raw fitness *
//as well as standardized fitness for MultiPlexer Problems               *
//Programmed by NXH.                                                     *
//************************************************************************
#ifndef __MULTIPLEXER__
#define __MULTIPLEXER__

#include<iostream.h>
#include<string.h>
#include<stdio.h>
#include<assert.h>
#include<math.h>
#include<stdlib.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
#include "dtree.h"
#include  "genetic.h"
#define NUMADDRESS  2
#define NUMVAR  6
#define NUMFITCASE 64
struct _sample{
		 BOOL x[NUMVAR];
		 BOOL y;
		};
typedef struct _sample sample;
sample fitcase[NUMFITCASE];
BOOL SuccPredicate=FALSE;
void SetFitCase()
{
 int i,h,l,t;

 for(i=0;i<NUMFITCASE;i++)
  {
   //change to boonlean
   h=i;
   for(l=0;l<NUMVAR;l++)
    {
     if((h % 2)==1)
       fitcase[i].x[l]=TRUE;
     else
      fitcase[i].x[l]=FALSE;
     h=h/2;
    }
  //compute the address value
   h=0;
   t=1;
   for(l=0;l<NUMADDRESS;l++)
    {
     if(fitcase[i].x[l])
       h=h+t;
     t=t*2;
    }
    fitcase[i].y=fitcase[i].x[h+NUMADDRESS];
  }
}
//Some Boolean function
BOOL IF(BOOL a, BOOL b, BOOL c)
{
 if(a)
  return b;
 else
 return c;
}
BOOL AND(BOOL a, BOOL b)
{
 if (a&&b)
  return TRUE;
 else
  return FALSE;
}
BOOL OR(BOOL a,BOOL b)
{
 if(a||b)
  return TRUE;
 else
  return FALSE;
}
BOOL NOT(BOOL a)
{
 if(!a)
  return TRUE;
 else
 return FALSE;
}
//********************************************************************
//This function compute the function (in tree form) with given input *
//********************************************************************
BOOL Compute(node *st,BOOL x[NUMVAR])
{
 node *p,*q;
 BOOL l,r;
 if (!strcmp(st->name,"a0"))
     return x[0];
  else
    if (!strcmp(st->name,"a1"))
      return x[1];
   else
    if (!strcmp(st->name,"d0"))
     return x[2];
   else
   if (!strcmp(st->name,"d1"))
     return x[3];
   else
     if (!strcmp(st->name,"d2"))
      return x[4];
     else
     if (!strcmp(st->name,"d3"))
      return x[5];
 else //st->name="BOOL"
  {
   p=st->children;
   q=p->sibling;
   if(!strcmp(p->name,"NOT"))//not
    return NOT(Compute(q,x));
    else
     if(!strcmp(p->name,"IF"))
      return IF(Compute(q,x),Compute(q->sibling,x),Compute(q->sibling->sibling,x));
     else
      {
       l=Compute(p,x);
       if(q!=NULL)
	{
	 r=Compute(q->sibling,x);
	 if(!strcmp(q->name,"AND"))
	    return AND(l,r);
	 else //OR
	    return OR(l,r);
	}
       else
	return l;
      }
  }
}
//*********************************************
//Compute the raw fitness for an invidual     *
//*********************************************
double ComputeRF(stree st)
{
  double sum=0;
  BOOL t;
  int i;
  int hit;
  hit=0;
  for(i=0;i<NUMFITCASE;i++)
   {
     t=Compute(st.root,fitcase[i].x);
     if(t==fitcase[i].y)
     {
      hit++;
      sum++;
     }
   }
   if(hit==NUMFITCASE)
     SuccPredicate=TRUE;
   return sum;
}
//**************************************
//This function build the vocabulary   *
//**************************************
void BuildVocab(void)
{
 char *lx[]={"a0","a1","d0","d1","d2","d3"};
 int i,j;
 stree s;
 vocab=new lexicon[3];
 lexnum=3;
 strcpy(vocab[0].name,"T1");
 strcpy(vocab[1].name,"T2");
 strcpy(vocab[2].name,"T3");
 for(i=0;i<lexnum;i++)//add lexeme
  {
    vocab[i].card=NUMVAR;
    for(j=0;j<NUMVAR;j++)
    {
     s=InitTree(lx[j]);
     vocab[i].lexemes[j].lex=s.root;
  //  vocab[i].lexemes[j].prob=1.0;
    if(i==2)
    {
     if(j<NUMADDRESS)
       vocab[i].lexemes[j].prob=2.0;//uniform
     else
       vocab[i].lexemes[j].prob=1.0;
    }
   else
     if(i==0)
      {
       if(j<NUMADDRESS)
	vocab[i].lexemes[j].prob=16.0;
       else
	vocab[i].lexemes[j].prob=1.0;
      }
     else //i=1
      {
       if(j>=NUMADDRESS)
	vocab[i].lexemes[j].prob=4.0;
	 else
	vocab[i].lexemes[j].prob=1.0;
      }
    }
  }
 NormaliseVD();
}
//****************************************************************
//This function initializes elementary trees and adjoining table *
//****************************************************************
void LoadEtree()
{
 cout<<endl<<"Loading elementary trees...";
 etree=new stree[14];assert(etree!=NULL);
 etree[0]=InitTree("BOOL(a0)");
 etree[1]=InitTree("BOOL(a1)");
 etree[2]=InitTree("BOOL(d0)");
 etree[3]=InitTree("BOOL(d1)");
 etree[4]=InitTree("BOOL(d2)");
 etree[5]=InitTree("BOOL(d3)");

 etree[6]=InitTree("BOOL(IF,BOOL(T1),BOOL(T2),BOOL~)");//IF Trees
 etree[7]=InitTree("BOOL(IF,BOOL(T1),BOOL~,BOOL(T2))");
 etree[8]=InitTree("BOOL(IF,BOOL~,BOOL(T2),BOOL(T2))");

 etree[9]=InitTree("BOOL(BOOL(T3),AND,BOOL~)"); //AND trees
 etree[10]=InitTree("BOOL(BOOL~,AND,BOOL(T3))");

 etree[11]=InitTree("BOOL(BOOL(T3),OR,BOOL~)");//OR tree
 etree[12]=InitTree("BOOL(BOOL~,OR,BOOL(T3))");

 etree[13]=InitTree("BOOL(NOT,BOOL~)");

 numetree=14;
 numalpha=6;
 numbeta=8;
 cout<<endl<<"Done!";
 cout<<endl<<"Building Vocabulary...";
 BuildVocab();
 cout<<endl<<"Done!";

 cout<<endl<<"Buiding adjoining table...";
 BuildAdTable();
 cout<<endl<<"Done!"<<entrynum;
}
//************************************************************
//This function transforms a derivation tree to a parse tree *
//************************************************************
node *Stree2ParseTree(node *st)
{
  node *p,*q,*t,*t1,*t2,*t3;
 if(!strcmp(st->name,"BOOL"))
  {
   p=st->children;
   q=p->sibling;
   if(!strcmp(p->name,"NOT"))//not
   {
    t=MakeNode("NOT");
    t->children=Stree2ParseTree(q);
    return t;
   }
    else
     if(!strcmp(p->name,"IF"))
     {
      t=MakeNode("IF");
      t1=Stree2ParseTree(q);
      t2=Stree2ParseTree(q->sibling);
      t3=Stree2ParseTree(q->sibling->sibling);
      t->children=t1;
      t1->sibling=t2;
      t2->sibling=t3;
      t3->sibling=NULL;
      return t;
     }
     else
       if(q!=NULL)
	{
	 t1=Stree2ParseTree(p);
	 t2=Stree2ParseTree(q->sibling);
	 if(!strcmp(q->name,"AND"))
	    t=MakeNode("AND");
	 else //OR
	    t=MakeNode("OR");
	 t->children=t1;
	 t1->sibling=t2;
	 t2->sibling=NULL;
	 return t;
	}
       else
	return Stree2ParseTree(p);
    }
 else
  {
    t=MakeNode(st->name);
    return t;
  }
}

#endif
