#ifndef __SPEA2__
#define __SPEA2__
//This file contain SPEA2 implementation for multi-objective optimization
//Programmed by NXH
//#include<conio.h>//Delete this line if you are in Unix
#include<iostream.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<math.h>
#include<assert.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
#include  "dtree.h"
#include "genetic.h"
//#include "mulplex.h"
#include "symreg.h"
BOOL strongdominance;
FILE *fppop,*fppar,*fpstat;
//**************************************
//Compute the fitness of an individual *
//**************************************
double ComputeIndividualRF(individual v)
{
 stree s;
 double f;
 s=Decode(v,FALSE);
 f=ComputeRF(s);
 DeleteTree(s);
 return f;
}
//********************************************************************
//Marking and pruning non-fitness contributing parts of an individual*
//********************************************************************
SIZE _activesize;
double _variance,_rawfitness;
void OldMarkingDtree1(dnode *t,individual &v) //use for sub-code of type 1 only (i.e complete subtree)
{
 stree s;
 double f;
 dnode *p;
 SIZE size1;
 if(t)
 {
  //compute the size
  size1=GetSizeDNode(t,TRUE);
  if(_activesize-size1>=MINSIZE)//only do the marking if the result bigger than MIN SIZE
   {
    t->id=-t->id;//temporary marked off
    f=ComputeIndividualRF(v); //compute the new fitness

   // cout<<endl<<"Fitness:"<<f;
   // cout<<endl<<"Active size:"<<_activesize;
    if(fabs(f-_rawfitness)<= _variance) //if the change is insignificant
       _activesize-=size1;
    else
      {
       t->id=-t->id;
       //try subtrees
       p=t->children;
      while(p)
       {
	OldMarkingDtree1(p,v);
	p=p->sibling;
       }
      }
   }

  else
   {
    p=t->children;
    while(p)
     {
       OldMarkingDtree1(p,v);
       p=p->sibling;
     }
   }
 }
}
void MarkingDtree1(dnode *t,individual &v) //use for both types of sub-code added in version 3.1
{
 stree s;
 double f;
 dnode *p;
 SIZE size1,size2;
 BOOL type1OK, type2OK;
 if(t)
 {
  //compute the size
  size1=GetSizeDNode(t,TRUE);
  //first try the subcode of type1
 type1OK=FALSE;
 type2OK=FALSE;
 if(_activesize-size1>=MINSIZE)//only do the marking if the result bigger than MIN SIZE
   {
    t->id=-t->id;//temporary marked off
    f=ComputeIndividualRF(v); //compute the new fitness
    if(fabs(f-_rawfitness)<= _variance) //if the change is insignificant
     {
     _activesize-=size1;
      type1OK=TRUE;
     }
    else
      t->id=-t->id; //turns it back on
    }
 if(!type1OK) //if type1 subcode is not marked then try type2
 { 
  if(t!=v.chrom) //not the root
   if(RootAdjoined(t))
    {
     p=t->children;
     while(p->sibling) p=p->sibling;
     size2=GetSizeDNode(p,TRUE);
     size2=size1-size2;
     if(_activesize-size2>=MINSIZE)
      {
       t->id=-(t->id+numetree); //temporarily marked for subcode type 2
       f=ComputeIndividualRF(v);
       if(fabs(f-_rawfitness)<=_variance)//if the change is insinificant
	{
         _activesize-=size2;
         type2OK=TRUE;
        }
       else
	t->id=-(t->id+numetree);
      }
    } 
     
//try the children
   if(!type2OK) //
   {
    p=t->children;
    while(p)
     {
       MarkingDtree1(p,v);
       p=p->sibling;
     }
    }
   else
    {
     p=t->children;
     while(p->sibling) p=p->sibling;
     MarkingDtree1(p,v);
     }
  }

 }
}
void OldDeleteIntron(individual &v) //delete introns as subcode of type 1 (complete sub-tree) only
{
 _variance=0.0;//subject to change
 _activesize=v.size;
_rawfitness=ComputeIndividualRF(v);
 OldMarkingDtree1(v.chrom,v);
// cout <<endl<<"Active size:"<<_activesize;
 PruneDNode(v.chrom,NULL);
 v.size=_activesize;
 v.branch=GetAVGDNode(v.chrom);
}
void DeleteIntron(individual &v) //delete introns as subcodes of type 1 and 2
{
 _variance=0.0;//subject to change
 _activesize=v.size;
_rawfitness=ComputeIndividualRF(v);
 MarkingDtree1(v.chrom,v);
// cout <<endl<<"Active size:"<<_activesize;
 PruneDNode(v.chrom,NULL);
 v.size=_activesize;
 v.branch=GetAVGDNode(v.chrom);
}

//*******************************************************************************
//Marking and extracting the most effective (destructive) part of an individual *
//*******************************************************************************
dnode *_subcode;
double _rawfitness2,_mloss;
BOOL  _effordes,_subcodetype;
void OldMarkingDtree2(dnode *t,dnode *parent,individual &v) //marking subcode of type 1 only
{
 stree s;
 double f;
 dnode *p;
 SIZE size1;
 if(parent) //excluding the root
 {
    t->id=-t->id;//temporary marked off
    f=ComputeIndividualRF(v); //compute the new fitness
     if(_effordes) //detect the most effective
      {   
       if(fabs(f-_rawfitness2)<_mloss)  
        {
         _subcode=t;
         _mloss=fabs(f-_rawfitness2);
	}
      }
     else
       if(fabs(f-_rawfitness2)>_mloss)
	{
	 _subcode=t;
	 _mloss=fabs(f-_rawfitness2);
        }
    t->id=-t->id;  
    p=t->children;
    while(p)
     {
       OldMarkingDtree2(p,t,v);
       p=p->sibling;
     }
  }
}
void MarkingDtree2(dnode *t,dnode *parent,individual  &v) //marking sub code of type1 and type 2
{
 stree s;
 double f;
 dnode *p;
 SIZE size1;
 if(parent) //excluding the root
 {
    t->id=-t->id;//temporary marked off
    f=ComputeIndividualRF(v); //compute the new fitness
     if(_effordes) //detect the most effective
      {   
       if(fabs(f-_rawfitness2)<_mloss ) 
        {
	 _subcode=t;
	 _mloss=fabs(f-_rawfitness2);
         _subcodetype=TRUE; //subcode type1
        }
      }
     else
       if(fabs(f-_rawfitness2)>_mloss)//detect the most destructive
        {
         _subcode=t;
	 _mloss=fabs(f-_rawfitness2);
          _subcodetype=TRUE; //subcode type1
	}
    t->id=-t->id;  
//try sub-code type 2
    if(RootAdjoined(t))
    {
     t->id=-(t->id+numetree);
     f=ComputeIndividualRF(v);
     if(_effordes)//detect the most effecitive 
       {
	if(fabs(f-_rawfitness2)<_mloss)
	 {
          _subcode=t;
          _mloss=fabs(f-_rawfitness2);
	  _subcodetype=FALSE;//subcode type2
	 }
       }
      else
      if(fabs(f-_rawfitness2)>_mloss) //dtect the most destructive
       {
	_subcode=t;
	_mloss=fabs(f-_rawfitness2);
	_subcodetype=FALSE;
       }
     t->id=-(t->id+numetree);
    }

    p=t->children;
    while(p)
     {
       MarkingDtree2(p,t,v);
       p=p->sibling;
     }
  }
}
dnode *OldSubCodeDetect(individual &v,BOOL type) //subcode of type 1 only
{
 if(type)//find the most effective subcode
  _mloss=999999999.99;
 else
  _mloss=-999999999.99;
 _rawfitness2=ComputeIndividualRF(v);
 _effordes=type;
 _subcode=NULL;
  OldMarkingDtree2(v.chrom,NULL,v);
 return _subcode;
}
dnode *SubCodeDetect(individual &v,BOOL type, BOOL &subcodetype) //subcode of type 1 and type 2
{
 if(type)//find the most effective subcode
  _mloss=999999999.99;
 else
  _mloss=-999999999.99;
 _rawfitness2=ComputeIndividualRF(v);
 _effordes=type;
 _subcode=NULL;
  MarkingDtree2(v.chrom,NULL,v);
 subcodetype=_subcodetype;
 return _subcode;
}

//********************************************************************
//     Functions for checking dominance and some general functions   *
//********************************************************************


//*******************************************************************
//Pareto domination comparision (the less the better)               *
//*******************************************************************
BOOL Pdominate(pindividual p,pindividual q)
{
 double test1,test2;
 test1=p.iv.fitness-q.iv.fitness;
 test2=p.sfitness-q.sfitness;
//strong dominance
if(strongdominance)
 {
  if(((test1<0)&&(test2<=0))||((test1<=0)&&(test2<0)))
    return TRUE;
   else
  return FALSE;
 }
 else
 {
  //weak dominance
  if((test1<=0)&&(test2<=0))
   return TRUE;
  else
   return FALSE;
  }
}
BOOL Pequal(pindividual p,pindividual q)
{
 if((p.iv.fitness==q.iv.fitness)&&(p.sfitness==q.sfitness))
  return TRUE;
 else
  return FALSE;
}
//********************************************
// Check if the i-th individual is dominated *
//********************************************
BOOL Dominated(int i,pindividual *p,int size)
{
  int j;
  for(j=0;j<size;j++)
   if(i!=j)
    if(Pdominate(p[j],p[i]))
     return TRUE;
 return FALSE;
}
BOOL NonSDominated(int i,pindividual *p,int size)
{
  int j;
  for(j=0;j<size;j++)
   if(Pdominate(p[j],p[i])&&!Pequal(p[i],p[j]))
     return FALSE;
 return TRUE;
}
//***********************************************
//Copy a pindividual                            *
//***********************************************
pindividual CopyPindividual(pindividual p,BOOL a)
{
 pindividual q;
 q.iv=CopyIndividual(p.iv,a);
 q.sfitness=p.sfitness;
 q.pfitness=p.pfitness;
// q.knb=p.knb;
 return q;
}
//*********************************************
//Function for pertubating the pdinviduals    *
//*********************************************
void Pertubate(pindividual *p,int size,int times)
{
 pindividual temp;
 int i,j,k;
 for(i=0;i<times;i++)
  {
    j=IRandom(0,size-1);
    k=IRandom(0,size-1);
    temp=CopyPindividual(p[j],FALSE);
    p[j]=CopyPindividual(p[k],FALSE);
    p[k]=CopyPindividual(temp,FALSE);
  }
}
//***************************************************************************
//This function return archive length for generation > 0; return 0 otherwise*
//***************************************************************************
int Alen(void)
{
 if(gen==0)
   return 0;
 else
  return alen;
}

//*************************************************************
//   Functions for computing the fitnesses of pdindividuals   *
//*************************************************************


//****************************************************
//Compute the euclide distance on the decision space *
//****************************************************
double Edistance(double a,double b,double c,double d)
{
 return sqrt((a-c)*(a-c)+(b-d)*(b-d));
}
//***********************************
// Sort functions for double number *
//***********************************
int sort_function( const void *a, const void *b)
{
    if(*(double *)a>*(double *)b)
     return 1;
    else
     if(*(double *)a==*(double *)b)
      return 0;
       else
	return -1;
}
//***************************************
//Compute the k-th nearest neigbours    *
//***************************************
double ComputeKN(pindividual q,pindividual *p,int size,int k)
{
 double *temp;
 int i;
 double kr;
 temp=new double[size]; assert(temp!=NULL);
for(i=0;i<size;i++)
  temp[i]=Edistance(q.iv.fitness,q.sfitness,p[i].iv.fitness,p[i].sfitness);
//sort the list
 qsort((void *)temp,size,sizeof(double), sort_function);
 kr=temp[k-1];
 delete [] temp;
 return kr;
}
//*********************************************
//compute the fitness type- 0 phenotype size, *
//1- phenotype height, 2- genotype size       *
//*********************************************
void ComputeFitness(int type)
{
 int i,j, pos;
 stree s,s1;
 individual t;
 double min,sum2,sum,sum3,min2;
 double *temp;
 //First Compute first fitness
 for(i=0;i<poplen;i++)
    oldpop[i].iv.fitness=ComputeIndividualRF(oldpop[i].iv);
 //Second Compute standardized first fitness
// for(i=0;i<poplen;i++)
//   oldpop[i].iv.fitness=(double) NUMFITCASE-oldpop[i].iv.fitness;
 //Compute the second objective fitness
 for(i=0;i<poplen;i++)
   switch (type)
    {
      case 0: //phenotype size;
	   s=Decode(oldpop[i].iv,FALSE);
	   s1.root=Stree2ParseTree(s.root);
	   s1.foot=NULL;
	   oldpop[i].sfitness=(double)GetSize(s1,TRUE);
	   DeleteTree(s);
	   DeleteTree(s1);
	   break;
     case 1: //phenotype height;
	   s=Decode(oldpop[i].iv,FALSE);
	   s1.root=Stree2ParseTree(s.root);
	   s1.foot=NULL;
	   oldpop[i].sfitness=(double)GetHeight(s1);
	   DeleteTree(s);
	   DeleteTree(s1);
	   break;
     case 2: //Genotype size;
	   oldpop[i].sfitness=(double)oldpop[i].iv.size;
	   break;

    }

 //Find the best of the current population and archive
   pos=0;
   min=oldpop[0].iv.fitness;
   for(i=1;i<poplen+Alen();i++)
    if (oldpop[i].iv.fitness<min)
    {
      min=oldpop[i].iv.fitness;
      pos=i;
    }
// if there are multiple appearance of min then use the second objective to choose
  min2=oldpop[pos].sfitness;
  for(i=0;i<poplen+Alen();i++)
    if(oldpop[i].iv.fitness==min)
     if(oldpop[i].sfitness<min2)
      {
       min2=oldpop[i].sfitness;
       pos=i;
      }
   sum=0;
   sum2=0;
   sum3=0;
   for(i=0;i<poplen;i++)
   {
    sum+=oldpop[i].iv.fitness;
    sum2+=oldpop[i].sfitness;
    sum3+=oldpop[i].iv.size;
   }
 //copy the best and average
   bestcurrent[gen]=CopyIndividual(oldpop[pos].iv,TRUE);
   averagefitness[gen]=sum/poplen;
   averagesfitness[gen]=sum2/poplen;
   averagesize[gen]=sum3/poplen;
//compute the pareto fitness of the population and archive

//first compute the raw pfitness
  temp=new double[poplen+Alen()];assert(temp!=NULL);
  for(i=0;i<poplen+Alen();i++)
  {
    oldpop[i].pfitness=0.0;
    temp[i]=0.0;
  }
 for(i=0;i<poplen+Alen();i++)
   for(j=0;j<poplen+Alen();j++)
    if(j!=i)
     if(Pdominate(oldpop[i],oldpop[j]))
      temp[i]++;
  for(i=0;i<poplen+Alen();i++)
    for(j=0;j<poplen+Alen();j++)
     if(j!=i)
     if(Pdominate(oldpop[j],oldpop[i]))
       oldpop[i].pfitness+=temp[j];
  delete [] temp;
//compute pfitness
 j=(int) sqrt(poplen+Alen());
for(i=0;i<poplen+Alen();i++)
 oldpop[i].pfitness=oldpop[i].pfitness+(1/(ComputeKN(oldpop[i],oldpop,poplen+Alen(),j)+2));
}

//***********************************************************************
//Some display functions for statistics, pop and p-frontier progression *
//***********************************************************************
//******************************
//Display the Pareto frontier  *
//******************************
void ParetoDisplay(FILE *fp)
{
 int i,j;
 j=0;
 for(i=poplen;i<poplen+alen;i++)
  if (NonSDominated(i-poplen,&newpop[poplen],alen))
     j++;
if(!fp)
 {
  cout<<endl<<"Number of points:"<<j<<endl;
  for(i=poplen;i<poplen+alen;i++)
   if (NonSDominated(i-poplen,&newpop[poplen],alen))
      cout<<"("<<newpop[i].iv.fitness<<","<<newpop[i].sfitness<<") ";
   //getch();
  }
 else
 {
  fprintf(fp,"%d\n",j);
  for(i=poplen;i<poplen+alen;i++)
   if (NonSDominated(i-poplen,&newpop[poplen],alen))
      fprintf(fp,"%lf %lf ",newpop[i].iv.fitness,newpop[i].sfitness);
  fprintf(fp,"\n");
 }
}
void PopulationDisplay(FILE *fp)
{
 int i;
// cout<<endl;
if(!fp)
 {
  for(i=0;i<poplen;i++)
   cout<<"("<<oldpop[i].iv.fitness<<","<<oldpop[i].sfitness<<","<<oldpop[i].pfitness<<") ";
  //getch();
 }
 else
 {
   for(i=0;i<poplen;i++)
    fprintf(fp,"%lf %lf ",oldpop[i].iv.fitness,oldpop[i].sfitness);
   fprintf(fp,"\n");
 }
}

//********************************************************************
// Some functions for the main evolution loop (SPEA2-implementation) *
//********************************************************************

//*******************************************************
//Tournament Selection on the population of Pindividual *
//*******************************************************
int PTour(int size)
{
   int i,j,pos;
   double min;
   assert(size<=MAXTOUR);
   pos=IRandom(poplen,alen+poplen-1);
   min=newpop[pos].pfitness;
   for(i=1;i<size;i++)
    {
      j=IRandom(poplen,alen+poplen-1);
      if (newpop[j].pfitness<min)
       {
	 min=newpop[j].pfitness;
	 pos=j;
       }
    }
   return pos;
}
//****************************************
//sort function for the pareto fitness   *
//****************************************
int sort_function2( const void *a, const void *b)
{
    if(((pindividual *)a)->pfitness>((pindividual *)b)->pfitness)
     return 1;
    else
     if(((pindividual *)a)->pfitness==((pindividual *)b)->pfitness)
      return 0;
       else
	return -1;
}

void Evolution()
{
 int i,j,l,k,mincount,pos,maxocc;
 double *knb;
 int *occs;
 pindividual *temp,parent1,parent2;
// stree s;
 double mink;
 BOOL dup;
 //Initialize a population
 RampInit(2,60,0.0);
  gen=0;
  //initialize for population and archive

 while(gen<generation)
 {
   ComputeFitness(0);
//   cout<<endl<<"generation..."<<gen;
//   ParetoDisplay1();
 //first copy all nondominated solutions
  PopulationDisplay(fppop);
   j=poplen+alen;
   for(i=0;i<poplen+Alen();i++)
    if(!Dominated(i,oldpop,poplen+Alen())) //non dominated solutions
      {
       j--;
       newpop[j]=CopyPindividual(oldpop[i],TRUE);
      }
//if the archive is not filled up
  if(j>poplen)
  {
   //choosing the rest from the oldpop and the archive
   //first sort the oldpop and old archive
 //  cout<<endl<<"Filling up:";
   qsort((void *)oldpop,poplen+Alen(),sizeof(pindividual), sort_function2);
 // for(i=0;i<poplen+Alen();i++)
 //   cout<<"("<<oldpop[i].iv.fitness<<","<<oldpop[i].sfitness<<","<<oldpop[i].pfitness<<") ";
 //  getch();
   //choosing the best of the rest
   i=0;
   while(!Dominated(i,oldpop,poplen+Alen())) i++;
   while(j>poplen)
     {
     //change here for the mixing of strong and weak dominance
      j--;
      newpop[j]=CopyPindividual(oldpop[i],TRUE);
      i++;
     }
  //  cout<<endl<<"Archive:"<<endl;
  //  for(i=poplen;i<poplen+Alen();i++)
  //  cout<<"("<<newpop[i].iv.fitness<<","<<newpop[i].sfitness<<","<<newpop[i].pfitness<<") ";
 //  getch();
   }
 if(j<poplen) //if it is over
  {
  // cout<<endl<<"Trunction";
   //eliminate the identical individuals first
   occs=new int[poplen+alen]; assert(occs!=NULL);
   for(i=j;i<poplen+alen;i++)
    occs[i]=1;
  for(i=j;i<poplen+alen;i++)
   if(occs[i]>0)
    for(k=i+1;k<poplen+alen;k++)
     if(Pequal(newpop[i],newpop[k]))
      {
       occs[i]++;
       occs[k]=0;
      }
   dup=TRUE;
  while(j<poplen && dup)
  {
   //find the maximum occurance
    maxocc=1;
    pos=j;
    for(i=j;i<poplen+alen;i++)
     if(occs[i]>maxocc)
      {
       maxocc=occs[i];
       pos=i;
      }
    if(maxocc==1)
     dup=FALSE;
    else
     {
      //randomly delelete the duplication
      k=IRandom(1,maxocc);
      //find the kth occurence
      i=j;
      while (k>0)
       {
	if(Pequal(newpop[i],newpop[pos]))
	 k--;
	if(k>0)
	 i++;
       }
     //delete that element
      if(i==pos)
	{
	//find next occurence of that individual
	 k=i+1;
	 while(!Pequal(newpop[k],newpop[i])) k++;
	  occs[k]=occs[i]-1;
	}
       else
	 occs[pos]--;
      //delete the element
       DeleteDNode(newpop[i].iv.chrom);
       for(k=i;k>j;k--)
       {
	newpop[k]=CopyPindividual(newpop[k-1],FALSE);
	occs[k]=occs[k-1];
       }
       j++;
     }
  }
 delete [] occs;
 if(j<poplen)
 {
  // cout<<endl<<"truncation knb"<<endl;
   knb=new double[poplen+alen]; assert(knb!=NULL);
   while(j<poplen)
   {
    //truncation
    k=2;
   for(i=j;i<poplen+alen;i++)
    knb[i]=0.0;
  //find the individual for elimination
   while(k<=poplen+alen-j)
    {
     //compute k-nearest neighbor
     for(i=j;i<poplen+alen;i++)
      if(knb[i]>=0)
       knb[i]=ComputeKN(newpop[i],&newpop[j],poplen+alen-j,k);
    //find the minimum
     mink=999999999;
     for(i=j;i<poplen+alen;i++)
      if(knb[i]>=0)
       if(knb[i]<mink)
	{
	 mink=knb[i];
	 pos=i;
	}
     //counting the min
      mincount=0;
      for(i=j;i<poplen+alen;i++)
       if(knb[i]==mink)
	mincount++;
       else
	 knb[i]=-1;//won't count in future
      if(mincount==1)
       k=alen+poplen-j+1;
      else
       k++;
    }
  DeleteDNode(newpop[pos].iv.chrom);
  for(i=pos;i>j;i--)
    newpop[i]=CopyPindividual(newpop[i-1],FALSE);
   j++;
  }
  delete [] knb;
 }
}
ParetoDisplay(fppar);
//Pertubate here
   l=0;
  while(l<poplen)
   {
    i=PTour(2);
    j=PTour(2);
    parent1=CopyPindividual(newpop[i],TRUE);
    parent2=CopyPindividual(newpop[j],TRUE);
   if(Flip(pcross))
   {
    if (SubTreeSwap(parent1.iv,parent2.iv,newpop[l].iv,newpop[l+1].iv,FALSE,0))
      ncross++;
    else
     {
      newpop[l]=CopyPindividual(parent1,FALSE);
      newpop[l+1]=CopyPindividual(parent2,FALSE);
     }
   }
    else
     {
      newpop[l]=CopyPindividual(parent1,FALSE);
      newpop[l+1]=CopyPindividual(parent2,FALSE);
     }
 if(Flip(preplace))
  if(ReplaceSubTree(newpop[l].iv,0,FALSE,TRUE,18))
    nreplace++;
  if(Flip(preplace))
   if(ReplaceSubTree(newpop[l+1].iv,0,FALSE,TRUE,18))
    nreplace++;
    l+=2;
   }
  for(k=0;k<poplen+Alen();k++)
   DeleteDNode(oldpop[k].iv.chrom);
  gen++;
  temp=oldpop;
  oldpop=newpop;
  newpop=temp;
 }
}
#endif
